import argparse
import json
import logging
import os
import sys
from collections import defaultdict

import networkx as nx
import pandas as pd
import stanza
from graphviz import Source
from networkx.algorithms.isomorphism import DiGraphMatcher
from xpotato.dataset.utils import default_pn_to_graph, ud_to_graph
from sklearn.metrics import precision_recall_fscore_support
from tqdm import tqdm
from tuw_nlp.grammar.text_to_4lang import TextTo4lang
from tuw_nlp.graph.utils import (
    GraphFormulaMatcher,
    GraphMatcher,
    graph_to_isi,
    graph_to_pn,
    pn_to_graph,
)
from tuw_nlp.text.pipeline import CachedStanzaPipeline


class GraphExtractor:
    def __init__(self, cache_dir=None, cache_fn=None, lang=None):
        self.cache_dir = cache_dir
        self.cache_fn = cache_fn
        self.lang = lang
        self.nlp = None
        self.matcher = None

    def init_nlp(self):
        if self.lang == "en_bio":
            nlp = stanza.Pipeline("en", package="craft")
        else:
            nlp = stanza.Pipeline(self.lang)
        self.nlp = CachedStanzaPipeline(nlp, self.cache_fn)

    def parse_iterable(self, iterable, graph_type="fourlang"):
        if graph_type == "fourlang":
            with TextTo4lang(
                lang=self.lang, nlp_cache=self.cache_fn, cache_dir=self.cache_dir
            ) as tfl:
                for sen in tqdm(iterable):
                    fl_graphs = list(tfl(sen))
                    g = fl_graphs[0]
                    for n in fl_graphs[1:]:
                        g = nx.compose(g, n)
                    yield g

        if graph_type == "ud":
            self.init_nlp()
            for sen in tqdm(iterable):
                doc = self.nlp(sen)
                g, _ = ud_to_graph(doc.sentences[0])
                for doc_sen in doc.sentences[1:]:
                    n, _ = ud_to_graph(doc_sen)
                    g = nx.compose(g, n)
                yield g


class FeatureEvaluator:
    def __init__(self, graph_format="ud"):
        self.graph_format = graph_format

    def match_features(self, dataset, features):
        graphs = dataset.graph.tolist()

        matches = []
        predicted = []

        matcher = GraphFormulaMatcher(features, converter=default_pn_to_graph)

        for i, g in tqdm(enumerate(graphs)):
            feats = matcher.match(g)
            for key, feature in feats:
                matches.append(features[feature])
                predicted.append(key)
                break
            else:
                matches.append("")
                predicted.append("")

        d = {
            "Sentence": dataset.text.tolist(),
            "Predicted label": predicted,
            "Matched rule": matches,
        }
        df = pd.DataFrame(d)
        return df

    def one_versus_rest(self, df, entity):
        mapper = {entity: 1}

        one_versus_rest_df = df.copy()
        one_versus_rest_df["one_versus_rest"] = [
            mapper[item] if item in mapper else 0 for item in df.label
        ]

        return one_versus_rest_df

    def rank_features(self, cl, features, orig_data, false_negatives):
        subset_data = orig_data.iloc[false_negatives]
        df, accuracy = self.evaluate_feature(cl, features, subset_data)

        features_stat = []

        for i, feature in enumerate(features):
            features_stat.append(
                (
                    feature,
                    df.iloc[i].Precision,
                    df.iloc[i].Recall,
                    df.iloc[i].Fscore,
                    df.iloc[i].Support,
                    len(df.iloc[i].True_positive_sens),
                    len(df.iloc[i].False_positive_sens)
                )
            )

        def rank(feature):
            return len(df.iloc[features.index(feature[0])].True_positive_graphs)

        return sorted(features_stat, key=rank, reverse=True)

    def train_feature(self, cl, feature, data, graph_format="ud"):
        feature_graph = default_pn_to_graph(feature)[0]

        graphs = data.graph.tolist()
        labels = self.one_versus_rest(data, cl).one_versus_rest.tolist()
        path = "trained_features.tsv"
        trained_features = []
        with open(path, "w+") as f:
            for i, g in enumerate(graphs):
                matcher = DiGraphMatcher(
                    g,
                    feature_graph,
                    node_match=GraphFormulaMatcher.node_matcher,
                    edge_match=GraphFormulaMatcher.edge_matcher,
                )
                if matcher.subgraph_is_isomorphic():
                    for iso_pairs in matcher.subgraph_isomorphisms_iter():
                        nodes = []
                        for k in iso_pairs:
                            if feature_graph.nodes[iso_pairs[k]]["name"] == ".*":
                                nodes.append(g.nodes[k]["name"])
                        if not nodes:
                            g2_to_g1 = {v: u for (u, v) in iso_pairs.items()}
                            for u, v, attrs in feature_graph.edges(data=True):
                                if attrs["color"] == ".*":
                                    edge = g.get_edge_data(g2_to_g1[u], g2_to_g1[v])[
                                        "color"
                                    ]
                                    nodes.append(edge)
                        nodes_str = ",".join(nodes)
                        label = labels[i]
                        sentence = data.iloc[i].text
                        f.write(f"{feature}\t{nodes_str}\t{sentence}\t{label}\n")
                        trained_features.append(
                            (feature, nodes_str, sentence, str(label))
                        )

        return self.cluster_feature(trained_features)

    def cluster_feature(self, trained_features):
        graphs = {}
        if os.path.isfile("longman_zero_paths_one_exp"):
            with open("longman_zero_paths_one_exp.json") as f:
                graphs = json.load(f)

        words = {}
        for fields in trained_features:
            words[fields[1] + "_" + fields[3]] = int(fields[3])
            feature = fields[0]
        graph = nx.MultiDiGraph()

        for word in words:
            if words[word] == 1:
                color = "green"
            else:
                color = "red"
            graph.add_node(word, color=color)
            word_clean = word.split("_")[0]
            if word_clean in graphs:
                hypernyms = graphs[word_clean]
                for hypernym in hypernyms:
                    hypernym_words = hypernyms[hypernym]
                    for w in hypernym_words:
                        if hypernym == "1":
                            graph.add_edge(word, w, color=hypernym)

        selected_words = self.select_words(trained_features)

        word_features = []

        word_features.append(feature.replace(".*", "|".join(selected_words)))

        return word_features

    def select_words(self, trained_features):
        features = []
        labels = []

        for fields in trained_features:
            features.append(fields[1])
            labels.append(int(fields[3]))
        words_to_measures = {
            word: {"TP": 0, "FP": 0, "TN": 0, "FN": 0} for word in set(features)
        }
        for word in words_to_measures:
            for i, label in enumerate(labels):
                if label and features[i] == word:
                    words_to_measures[word]["TP"] += 1
                if label and features[i] != word:
                    words_to_measures[word]["FN"] += 1
                if not label and features[i] == word:
                    words_to_measures[word]["FP"] += 1
                if not label and features[i] != word:
                    words_to_measures[word]["TN"] += 1

        for word in words_to_measures:
            TP = words_to_measures[word]["TP"]
            FP = words_to_measures[word]["FP"]
            TN = words_to_measures[word]["TN"]
            FN = words_to_measures[word]["FN"]

            precision = TP / (TP + FP)
            recall = TP / (TP + FN)

            words_to_measures[word]["precision"] = precision
            words_to_measures[word]["recall"] = recall

        selected_words = set()

        for word in words_to_measures:
            if words_to_measures[word]["precision"] > 0.9 and (
                words_to_measures[word]["TP"] > 1
                or words_to_measures[word]["recall"] > 0.01
            ):
                selected_words.add(word)

        return selected_words

    def evaluate_feature(self, cl, features, data, graph_format="ud"):
        measure_features = []
        graphs = data.graph.tolist()
        labels = self.one_versus_rest(data, cl).one_versus_rest.tolist()

        whole_predicted = []
        matched = defaultdict(list)

        # We want to view false negative examples for all rules, not rule specific
        false_neg_g = []
        false_neg_s = []
        false_neg_indices = []
        matcher = GraphFormulaMatcher(features, converter=default_pn_to_graph)
        for i, g in enumerate(graphs):
            feats = matcher.match(g)
            label = 0
            for key, feature in feats:
                matched[i].append(features[feature][0])
                label = 1
            whole_predicted.append(label)

            if label == 0 and labels[i] == 1:
                false_neg_g.append(g)
                sen = data.iloc[i].text
                lab = data.iloc[i].label
                false_neg_s.append((sen, lab))
                false_neg_indices.append(i)

        accuracy = []
        for pcf in precision_recall_fscore_support(
            labels, whole_predicted, average=None
        ):
            if len(pcf) > 1:
                accuracy.append(pcf[1])
            else:
                accuracy.append(0)

        for feat in features:
            measure = [feat[0]]
            false_pos_g = []
            false_pos_s = []
            true_pos_g = []
            true_pos_s = []
            predicted = []
            for i, g in enumerate(graphs):
                feats = matched[i]
                label = 1 if feat[0] in feats else 0
                if label == 1 and labels[i] == 0:
                    false_pos_g.append(g)
                    sen = data.iloc[i].text
                    lab = data.iloc[i].label
                    false_pos_s.append((sen, lab))
                if label == 1 and labels[i] == 1:
                    true_pos_g.append(g)
                    sen = data.iloc[i].text
                    lab = data.iloc[i].label
                    true_pos_s.append((sen, lab))
                predicted.append(label)
            for pcf in precision_recall_fscore_support(labels, predicted, average=None):
                if len(pcf) > 1:
                    measure.append(pcf[1])
                else:
                    measure.append(0)
            measure.append(false_pos_g)
            measure.append(false_pos_s)
            measure.append(true_pos_g)
            measure.append(true_pos_s)
            measure.append(false_neg_g)
            measure.append(false_neg_s)
            measure.append(false_neg_indices)
            measure.append(predicted)
            measure_features.append(measure)

        df = pd.DataFrame(
            measure_features,
            columns=[
                "Feature",
                "Precision",
                "Recall",
                "Fscore",
                "Support",
                "False_positive_graphs",
                "False_positive_sens",
                "True_positive_graphs",
                "True_positive_sens",
                "False_negative_graphs",
                "False_negative_sens",
                "False_negative_indices",
                "Predicted",
            ],
        )

        return df, accuracy



import argparse
import sys
import pandas as pd
import logging

from xpotato.dataset.hasoc_dataset import HasocDataset
from xpotato.dataset.semeval_dataset import SemevalDataset
from xpotato.graph_extractor.extract import GraphExtractor
from xpotato.models.model import GraphModel


def load_hasoc(path, output_graphs):
    df = pd.read_csv(path, delimiter="\t")
    data = HasocDataset(df)
    extractor = GraphExtractor(lang="en", cache_fn="en_nlp_cache")
    model = GraphModel()
    graphs = data.parse_graphs(extractor)
    data.set_graphs(graphs)
    data.graphs = graphs
    data.save_graphs(output_graphs)


def get_args():
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("-g", "--graph-type", type=str, default="fourlang")
    parser.add_argument("-d", "--dataset", type=str, default=None, required=True)
    parser.add_argument("-p", "--dataset-path", type=str, default=None, required=True)
    parser.add_argument("-of", "--output-graphs", type=str, default=None)
    parser.add_argument("-c", "--graphs-cache", type=str, default=None)

    return parser.parse_args()


def main():
    logging.basicConfig(
        level=logging.WARNING,
        format="%(asctime)s : "
        + "%(module)s (%(lineno)s) - %(levelname)s - %(message)s",
    )

    args = get_args()
    load_hasoc(args.dataset_path, args.output_graphs)


if __name__ == "__main__":
    main()



import argparse
import logging
import pickle
import sys
import json
from collections import defaultdict

import networkx as nx
import pandas as pd
from xpotato.dataset.utils import default_pn_to_graph
from xpotato.graph_extractor.extract import FeatureEvaluator


def get_args():
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("-t", "--graph-type", type=str, default="fourlang")
    parser.add_argument("-f", "--features", type=str, required=True)
    parser.add_argument("-d", "--dataset-path", type=str, default=None, required=True)

    return parser.parse_args()


def main():
    logging.basicConfig(
        level=logging.WARNING,
        format="%(asctime)s : "
        + "%(module)s (%(lineno)s) - %(levelname)s - %(message)s",
    )

    args = get_args()
    df = pd.read_pickle(args.dataset_path)

    with open(args.features) as f:
        features = json.load(f)

    feature_values = []
    for k in features:
        for f in features[k]:
            feature_values.append(f)
    evaluator = FeatureEvaluator()
    pred_df = evaluator.match_features(df, feature_values)
    if "label" in df:
        pred_df["label"] = df.label
    pred_df.to_csv(sys.stdout, sep="\t")


if __name__ == "__main__":
    main()



import argparse
import os
import time

from st_aggrid import AgGrid, GridOptionsBuilder, GridUpdateMode

from utils import *


def supervised_mode(
    evaluator, data, val_data, graph_format, feature_path, hand_made_rules
):
    if hand_made_rules:
        with open(hand_made_rules) as f:
            st.session_state.features = json.load(f)

    if not feature_path and not st.session_state.trained:
        st.sidebar.title("Train your dataset!")
        show_app = st.sidebar.button("Train")
        st.session_state.min_edge = st.sidebar.number_input(
            "Min edge in features", min_value=0, max_value=3, value=0, step=1
        )
        st.session_state.rank = st.sidebar.selectbox(
            "Rank features based on accuracy", options=[False, True]
        )
        if show_app:
            st.session_state.suggested_features = train_df(
                data, st.session_state.min_edge
            )
            st.session_state.trained = True
            with st_stdout("success"):
                print("Success, your dataset is trained, wait for the app to load..")
                time.sleep(3)
                rerun()
        st.markdown(
            "<h3 style='text-align: center; color: black;'>Your dataset is shown below, click the train button to train your dataset!</h3>",
            unsafe_allow_html=True,
        )
        sample_df = AgGrid(data, width="100%", fit_columns_on_grid_load=True)

        st.write("label distribution:")
        st.bar_chart(data.groupby("label").size())

        st.write("sentence lenghts:")
        st.bar_chart(data.text.str.len())

        st.write("common words:")
        st.bar_chart(
            pd.Series(" ".join(data["text"]).lower().split()).value_counts()[:100]
        )

    if st.session_state.trained or feature_path:
        col1, col2 = st.columns(2)

        if (
            feature_path
            and os.path.exists(feature_path)
            and not st.session_state.suggested_features
        ):
            with open(feature_path) as f:
                st.session_state.suggested_features = json.load(f)

        if not st.session_state.features:
            for key in st.session_state.suggested_features:
                pop_len = (
                    5
                    if len(st.session_state.suggested_features[key]) > 5
                    else len(st.session_state.suggested_features[key])
                )
                st.session_state.features[key] = [
                    st.session_state.suggested_features[key].pop(0)
                    for _ in range(pop_len)
                ]

        col1.header("Rule to apply")

        col2.header("Graphs and results")

        # if graph_format == "fourlang":
        #    tfl = load_text_to_4lang()

        with col1:
            classes = st.selectbox(
                "Choose class", list(st.session_state.features.keys())
            )

            st.session_state.feature_df = get_df_from_rules(
                [";".join(feat[0]) for feat in st.session_state.features[classes]],
                [";".join(feat[1]) for feat in st.session_state.features[classes]],
            )

            with st.form("example form") as f:
                gb = GridOptionsBuilder.from_dataframe(st.session_state.feature_df)
                # make all columns editable
                gb.configure_columns(["rules", "negated_rules"], editable=True)
                gb.configure_selection(
                    "multiple",
                    use_checkbox=True,
                    groupSelectsChildren=True,
                    groupSelectsFiltered=True,
                    # ◙pre_selected_rows=[1,2]
                )
                go = gb.build()
                ag = AgGrid(
                    st.session_state.feature_df,
                    gridOptions=go,
                    key="grid1",
                    allow_unsafe_jscode=True,
                    reload_data=True,
                    update_mode=GridUpdateMode.MODEL_CHANGED
                    | GridUpdateMode.VALUE_CHANGED,
                    width="100%",
                    theme="material",
                    fit_columns_on_grid_load=True,
                )

                delete_or_train = st.radio(
                    "Delete or Train selected rules", ("none", "delete", "train")
                )
                submit = st.form_submit_button(label="save updates")
                evaluate = st.form_submit_button(label="evaluate selected")

            if evaluate:
                feature_list = []
                selected_rules = (
                    ag["selected_rows"]
                    if ag["selected_rows"]
                    else ag["data"].to_dict(orient="records")
                )
                for rule in selected_rules:
                    positive_rules = (
                        rule["rules"].split(";")
                        if "rules" in rule and rule["rules"].strip()
                        else []
                    )
                    negated_rules = (
                        rule["negated_rules"].split(";")
                        if "negated_rules" in rule and rule["negated_rules"].strip()
                        else []
                    )
                    feature_list.append(
                        [
                            positive_rules,
                            negated_rules,
                            classes,
                        ]
                    )
                st.session_state.sens = [";".join(feat[0]) for feat in feature_list]
                with st.spinner("Evaluating rules..."):
                    (
                        st.session_state.df_statistics,
                        st.session_state.whole_accuracy,
                    ) = evaluator.evaluate_feature(
                        classes, feature_list, data, graph_format
                    )
                    (
                        st.session_state.val_dataframe,
                        st.session_state.whole_accuracy_val,
                    ) = evaluator.evaluate_feature(
                        classes,
                        feature_list,
                        val_data,
                        graph_format,
                    )
                st.success("Done!")
                rerun()

            if submit:
                delete = delete_or_train == "delete"
                train = delete_or_train == "train"

                st.session_state.rows_to_delete = [
                    r["rules"] for r in ag["selected_rows"]
                ]
                st.session_state.rls_after_delete = []

                negated_list = ag["data"]["negated_rules"].tolist()
                feature_list = []
                for i, rule in enumerate(ag["data"]["rules"].tolist()):
                    if not negated_list[i].strip():
                        feature_list.append([rule.split(";"), [], classes])
                    else:
                        feature_list.append(
                            [
                                rule.split(";"),
                                negated_list[i].strip().split(";"),
                                classes,
                            ]
                        )
                if st.session_state.rows_to_delete and delete:
                    for r in feature_list:
                        if ";".join(r[0]) not in st.session_state.rows_to_delete:
                            st.session_state.rls_after_delete.append(r)
                elif st.session_state.rows_to_delete and train:
                    st.session_state.rls_after_delete = copy.deepcopy(feature_list)
                    rule_to_train = st.session_state.rows_to_delete[0]
                    if ";" in rule_to_train or ".*" not in rule_to_train:
                        st.text("Only single and underspecified rules can be trained!")
                    else:
                        selected_words = evaluator.train_feature(
                            classes, rule_to_train, data, graph_format
                        )

                        for f in selected_words:
                            st.session_state.rls_after_delete.append([[f], [], classes])
                else:
                    st.session_state.rls_after_delete = copy.deepcopy(feature_list)

                if st.session_state.rls_after_delete and not delete:
                    save_after_modify(hand_made_rules, classes)

            if st.session_state.rows_to_delete and delete_or_train == "delete":
                with st.form("Delete form"):
                    st.write("The following rules will be deleted, do you accept it?")
                    st.write(st.session_state.rows_to_delete)
                    save_button = st.form_submit_button("Accept Delete")

                if save_button:
                    save_after_modify(hand_made_rules, classes)

            add_rule_manually(classes, hand_made_rules)
            rank_and_suggest(classes, data, evaluator)

            if st.session_state.ml_feature:
                show_ml_feature(classes, hand_made_rules)

        with col2:
            if not st.session_state.df_statistics.empty and st.session_state.sens:
                if st.session_state.sens:
                    nodes, option = rule_chooser()
                st.markdown(
                    f"<span>Result of using all the rules: Precision: <b>{st.session_state.whole_accuracy[0]:.3f}</b>, \
                        Recall: <b>{st.session_state.whole_accuracy[1]:.3f}</b>, Fscore: <b>{st.session_state.whole_accuracy[2]:.3f}</b>, \
                            Support: <b>{st.session_state.whole_accuracy[3]}</b></span>",
                    unsafe_allow_html=True,
                )
                (
                    fn_graphs,
                    fn_sentences,
                    fp_graphs,
                    fp_sentences,
                    fscore,
                    prec,
                    predicted,
                    recall,
                    support,
                    tp_graphs,
                    tp_sentences,
                ) = extract_data_from_dataframe(option)

                st.markdown(
                    f"<span>The rule's result: Precision: <b>{prec:.3f}</b>, Recall: <b>{recall:.3f}</b>, \
                        Fscore: <b>{fscore:.3f}</b>, Support: <b>{support}</b></span>",
                    unsafe_allow_html=True,
                )

                with st.expander("Show validation data", expanded=False):
                    val_prec = st.session_state.val_dataframe.iloc[
                        st.session_state.sens.index(option)
                    ].Precision
                    val_recall = st.session_state.val_dataframe.iloc[
                        st.session_state.sens.index(option)
                    ].Recall
                    val_fscore = st.session_state.val_dataframe.iloc[
                        st.session_state.sens.index(option)
                    ].Fscore
                    val_support = st.session_state.val_dataframe.iloc[
                        st.session_state.sens.index(option)
                    ].Support
                    st.markdown(
                        f"<span>Result of using all the rules on the validation data: Precision: <b>{st.session_state.whole_accuracy_val[0]:.3f}</b>, \
                            Recall: <b>{st.session_state.whole_accuracy_val[1]:.3f}</b>, Fscore: <b>{st.session_state.whole_accuracy_val[2]:.3f}</b>, \
                                Support: <b>{st.session_state.whole_accuracy_val[3]}</b></span>",
                        unsafe_allow_html=True,
                    )
                    st.markdown(
                        f"<span>The rule's result on the validation data: Precision: <b>{val_prec:.3f}</b>, \
                            Recall: <b>{val_recall:.3f}</b>, Fscore: <b>{val_fscore:.3f}</b>, \
                                Support: <b>{val_support}</b></span>",
                        unsafe_allow_html=True,
                    )

                tp_fp_fn_choice = (
                    "True Positive graphs",
                    "False Positive graphs",
                    "False Negative graphs",
                )
                tp_fp_fn = st.selectbox(
                    "Select the graphs you want to view", tp_fp_fn_choice
                )
                if tp_fp_fn == "False Positive graphs":
                    if fp_graphs:
                        graph_viewer("FP", fp_graphs, fp_sentences, nodes)

                elif tp_fp_fn == "True Positive graphs":
                    if tp_graphs:
                        graph_viewer("TP", tp_graphs, tp_sentences, nodes)

                elif tp_fp_fn == "False Negative graphs":
                    if fn_graphs:
                        graph_viewer("FN", fn_graphs, fn_sentences, nodes)


def unsupervised_mode(
    evaluator, train_data, graph_format, feature_path, hand_made_rules
):
    data = read_train(train_data)
    if hand_made_rules:
        with open(hand_made_rules) as f:
            st.session_state.features = json.load(f)
    if "df" not in st.session_state:
        st.session_state.df = data.copy()
        if "annotated" not in st.session_state.df:
            st.session_state.df["annotated"] = False
        if "applied_rules" not in st.session_state.df:
            st.session_state.df["applied_rules"] = [
                [] for _ in range(len(st.session_state.df))
            ]
        if "index" not in st.session_state.df:
            st.session_state.df.reset_index(level=0, inplace=True)

    df_annotated = st.session_state.df[st.session_state.df.annotated == True][
        ["index", "text", "label", "applied_rules"]
    ]
    df_unannotated = st.session_state.df[st.session_state.df.annotated == False][
        ["index", "text", "label", "applied_rules"]
    ]

    if "labels" not in st.session_state:
        st.text("Before we start, please provide labels you want to train")
        user_input = st.text_input("label encoding", "NOT:0,OFF:1")

        st.session_state.labels = {
            label.split(":")[0]: int(label.split(":")[1])
            for label in user_input.split(",")
        }

        st.write(st.session_state.labels)
        st.session_state.inverse_labels = {
            v: k for (k, v) in st.session_state.labels.items()
        }
    else:
        st.markdown(
            f"<span><b>Annotate samples here:</b></span>",
            unsafe_allow_html=True,
        )

        if st.session_state.applied_rules:
            st.markdown(
                f"<span>Currently the following rules are applied:</span>",
                unsafe_allow_html=True,
            )
            st.write(st.session_state.applied_rules)
        with st.form("annotate form") as f:
            gb = GridOptionsBuilder.from_dataframe(df_unannotated)
            gb.configure_default_column(
                editable=True,
                resizable=True,
                sorteable=True,
                wrapText=True,
                autoHeight=True,
            )
            # make all columns editable
            gb.configure_selection(
                "multiple",
                use_checkbox=True,
                groupSelectsChildren=True,
                groupSelectsFiltered=True,
            )
            go = gb.build()
            ag = AgGrid(
                df_unannotated,
                gridOptions=go,
                key="grid2",
                allow_unsafe_jscode=True,
                reload_data=True,
                update_mode=GridUpdateMode.MODEL_CHANGED | GridUpdateMode.VALUE_CHANGED,
                width="100%",
                theme="material",
                fit_columns_on_grid_load=True,
            )

            annotate = st.form_submit_button("Annotate")

        if annotate:
            if ag["selected_rows"]:
                for row in ag["selected_rows"]:
                    st.session_state.df.loc[
                        row["index"], "label"
                    ] = st.session_state.inverse_labels[1]
                    st.session_state.df.loc[row["index"], "annotated"] = True
                save_dataframe(st.session_state.df, train_data)
                rerun()

        st.markdown(
            f"<span>Samples you have already annotated:</span>",
            unsafe_allow_html=True,
        )
        with st.form("annotated form") as f:
            gb = GridOptionsBuilder.from_dataframe(df_annotated)
            gb.configure_default_column(
                editable=True,
                resizable=True,
                sorteable=True,
                wrapText=True,
            )
            # make all columns editable
            gb.configure_selection(
                "multiple",
                use_checkbox=True,
                groupSelectsChildren=True,
                groupSelectsFiltered=True,
            )
            go = gb.build()
            ag_ann = AgGrid(
                df_annotated,
                gridOptions=go,
                key="grid3",
                allow_unsafe_jscode=True,
                reload_data=True,
                update_mode=GridUpdateMode.MODEL_CHANGED | GridUpdateMode.VALUE_CHANGED,
                width="100%",
                theme="material",
                fit_columns_on_grid_load=True,
            )

            clear_annotate = st.form_submit_button("Clear annotation")

        if clear_annotate:
            if ag_ann["selected_rows"]:
                for row in ag_ann["selected_rows"]:
                    st.session_state.df.loc[
                        row["index"], "label"
                    ] = st.session_state.inverse_labels[1]
                    st.session_state.df.loc[row["index"], "annotated"] = False
                    st.session_state.df.loc[row["index"], "label"] = ""
                save_dataframe(st.session_state.df, train_data)
                rerun()

        train = st.button("Train!")
        st.session_state.min_edge = st.sidebar.number_input(
            "Min edge in features", min_value=0, max_value=3, value=0, step=1
        )
        st.session_state.rank = st.sidebar.selectbox(
            "Rank features based on accuracy", options=[False, True]
        )
        if train:
            df_to_train = st.session_state.df.copy()
            df_to_train = df_to_train[df_to_train.applied_rules.map(len) == 0]

            if not df_to_train.empty:
                st.session_state.trained = True
                df_to_train["label"] = df_to_train["label"].apply(
                    lambda x: st.session_state.inverse_labels[0] if not x else x
                )
                df_to_train["label_id"] = df_to_train["label"].apply(
                    lambda x: st.session_state.labels[x]
                )

                positive_size = df_to_train.groupby("label").size()[
                    st.session_state.inverse_labels[1]
                ]
                df_to_train = df_to_train.groupby("label").sample(
                    n=positive_size, random_state=1, replace=True
                )
                st.session_state.suggested_features = train_df(
                    df_to_train, st.session_state.min_edge, st.session_state.rank
                )
                st.session_state.df_to_train = df_to_train
                st.session_state.df_statistics = pd.DataFrame
                for key in st.session_state.suggested_features:
                    if key not in st.session_state.features:
                        st.session_state.features[key] = [
                            st.session_state.suggested_features[key].pop(0)
                        ]
                    else:
                        st.session_state.features[key].append(
                            st.session_state.suggested_features[key].pop(0)
                        )

            else:
                st.write("Empty dataframe!")

        col1, col2 = st.columns(2)

        if st.session_state.trained and st.session_state.suggested_features:
            with col1:

                if not st.session_state.features:
                    for key in st.session_state.suggested_features:
                        st.session_state.features[key] = [
                            st.session_state.suggested_features[key].pop(0)
                        ]

                classes = st.selectbox(
                    "Choose class", list(st.session_state.features.keys())
                )

                st.session_state.feature_df = get_df_from_rules(
                    [";".join(feat[0]) for feat in st.session_state.features[classes]],
                    [";".join(feat[1]) for feat in st.session_state.features[classes]],
                )

                with st.form("example form") as f:
                    gb = GridOptionsBuilder.from_dataframe(st.session_state.feature_df)
                    # make all columns editable
                    gb.configure_columns(["rules", "negated_rules"], editable=True)
                    gb.configure_selection(
                        "multiple",
                        use_checkbox=True,
                        groupSelectsChildren=True,
                        groupSelectsFiltered=True,
                        # ◙pre_selected_rows=[1,2]
                    )
                    go = gb.build()
                    ag = AgGrid(
                        st.session_state.feature_df,
                        gridOptions=go,
                        key="grid1",
                        allow_unsafe_jscode=True,
                        reload_data=True,
                        update_mode=GridUpdateMode.MODEL_CHANGED
                        | GridUpdateMode.VALUE_CHANGED,
                        width="100%",
                        theme="material",
                        fit_columns_on_grid_load=True,
                    )

                    delete_or_train = st.radio(
                        "Delete or Train selected rules", ("none", "delete", "train")
                    )
                    submit = st.form_submit_button(label="save updates")
                    evaluate = st.form_submit_button(label="evaluate selected")
                    annotate = st.form_submit_button(label="annotate based on selected")

                feature_list = []
                selected_rules = (
                    ag["selected_rows"]
                    if ag["selected_rows"]
                    else ag["data"].to_dict(orient="records")
                )
                for rule in selected_rules:
                    positive_rules = (
                        rule["rules"].split(";")
                        if "rules" in rule and rule["rules"].strip()
                        else []
                    )
                    negated_rules = (
                        rule["negated_rules"].split(";")
                        if "negated_rules" in rule and rule["negated_rules"].strip()
                        else []
                    )
                    feature_list.append(
                        [
                            positive_rules,
                            negated_rules,
                            classes,
                        ]
                    )

                if evaluate or annotate:
                    st.session_state.sens = [";".join(feat[0]) for feat in feature_list]
                    with st.spinner("Evaluating rules..."):
                        (
                            st.session_state.df_statistics,
                            st.session_state.whole_accuracy,
                        ) = evaluator.evaluate_feature(
                            classes,
                            feature_list,
                            st.session_state.df,
                            graph_format,
                        )

                    st.success("Done!")

                    if annotate:
                        predicted_rules = [[] for _ in range(len(st.session_state.df))]
                        st.session_state.applied_rules = st.session_state.sens
                        for j, opt in enumerate(st.session_state.sens):
                            predicted = st.session_state.df_statistics.iloc[j].Predicted

                            predicted_indices = [
                                i for i, pred in enumerate(predicted) if pred == 1
                            ]

                            for ind in predicted_indices:
                                predicted_rules[ind].append(opt)
                        annotate_df(predicted_rules)
                        st.session_state.trained = False

                    rerun()

                if submit:
                    delete = delete_or_train == "delete"
                    train = delete_or_train == "train"

                    st.session_state.rows_to_delete = [
                        r["rules"] for r in ag["selected_rows"]
                    ]
                    st.session_state.rls_after_delete = []

                    negated_list = ag["data"]["negated_rules"].tolist()
                    feature_list = []
                    for i, rule in enumerate(ag["data"]["rules"].tolist()):
                        if not negated_list[i].strip():
                            feature_list.append([rule.split(";"), [], classes])
                        else:
                            feature_list.append(
                                [
                                    rule.split(";"),
                                    negated_list[i].strip().split(";"),
                                    classes,
                                ]
                            )
                    if st.session_state.rows_to_delete and delete:
                        for r in feature_list:
                            if ";".join(r[0]) not in st.session_state.rows_to_delete:
                                st.session_state.rls_after_delete.append(r)
                    elif st.session_state.rows_to_delete and train:
                        st.session_state.rls_after_delete = copy.deepcopy(feature_list)
                        rule_to_train = st.session_state.rows_to_delete[0]
                        if ";" in rule_to_train or ".*" not in rule_to_train:
                            st.text(
                                "Only single and underspecified rules can be trained!"
                            )
                        else:
                            selected_words = evaluator.train_feature(
                                classes,
                                rule_to_train,
                                st.session_state.df,
                                graph_format,
                            )

                            for f in selected_words:
                                st.session_state.rls_after_delete.append(
                                    [[f], [], classes]
                                )
                    else:
                        st.session_state.rls_after_delete = copy.deepcopy(feature_list)

                    if st.session_state.rls_after_delete and not delete:
                        save_after_modify(hand_made_rules, classes)

                if st.session_state.rows_to_delete and delete_or_train == "delete":
                    with st.form("Delete form"):
                        st.write(
                            "The following rules will be deleted, do you accept it?"
                        )
                        st.write(st.session_state.rows_to_delete)
                        save_button = st.form_submit_button("Accept Delete")

                    if save_button:
                        save_after_modify(hand_made_rules, classes)

                add_rule_manually(classes, hand_made_rules)
                rank_and_suggest(
                    classes, st.session_state.df, evaluator, rank_false_negatives=False
                )

                if st.session_state.ml_feature:
                    show_ml_feature(classes, hand_made_rules)
            with col2:
                if not st.session_state.df_statistics.empty and st.session_state.sens:
                    if st.session_state.sens:
                        nodes, option = rule_chooser()
                    st.markdown(
                        f"<span>Result of using all the rules: Precision: <b>{st.session_state.whole_accuracy[0]:.3f}</b>, \
                            Recall: <b>{st.session_state.whole_accuracy[1]:.3f}</b>, Fscore: <b>{st.session_state.whole_accuracy[2]:.3f}</b>, \
                                Support: <b>{st.session_state.whole_accuracy[3]}</b></span>",
                        unsafe_allow_html=True,
                    )
                    (
                        fn_graphs,
                        fn_sentences,
                        fp_graphs,
                        fp_sentences,
                        fscore,
                        prec,
                        predicted,
                        recall,
                        support,
                        tp_graphs,
                        tp_sentences,
                    ) = extract_data_from_dataframe(option)

                    st.markdown(
                        f"<span>The rule's result: Precision: <b>{prec:.3f}</b>, Recall: <b>{recall:.3f}</b>, \
                            Fscore: <b>{fscore:.3f}</b>, Support: <b>{support}</b></span>",
                        unsafe_allow_html=True,
                    )

                    tp_fp_fn_choice = (
                        "Predicted",
                        "True Positive graphs",
                        "False Positive graphs",
                        "False Negative graphs",
                    )
                    tp_fp_fn = st.selectbox(
                        "Select the option you want to view", tp_fp_fn_choice
                    )

                    current_graph = None
                    if tp_fp_fn == "Predicted":
                        predicted_inds = [
                            i for i, pred in enumerate(predicted) if pred == 1
                        ]
                        if st.button("Previous Predicted"):
                            st.session_state.predicted_num = max(
                                0, st.session_state.predicted_num - 1
                            )
                        if st.button("Next Predicted"):
                            st.session_state.predicted_num = min(
                                st.session_state.predicted_num + 1,
                                len(predicted_inds) - 1,
                            )

                        if st.session_state.predicted_num > len(predicted_inds) - 1:
                            st.session_state.predicted_inds = 0

                        st.markdown(
                            f"<span><b>Sentence:</b> {st.session_state.df.iloc[predicted_inds[st.session_state.predicted_num]].text}</span>",
                            unsafe_allow_html=True,
                        )
                        st.markdown(
                            f"<span><b>Gold label:</b> {st.session_state.df.iloc[predicted_inds[st.session_state.predicted_num]].label}</span>",
                            unsafe_allow_html=True,
                        )
                        st.text(f"Predicted: {len(predicted_inds)}")
                        current_graph = st.session_state.df.iloc[
                            predicted_inds[st.session_state.predicted_num]
                        ].graph
                        st.graphviz_chart(
                            to_dot(
                                current_graph,
                                marked_nodes=set(nodes),
                            ),
                            use_container_width=True,
                        )

                    elif tp_fp_fn == "False Positive graphs":
                        if fp_graphs:
                            graph_viewer("FP", fp_graphs, fp_sentences, nodes)

                    elif tp_fp_fn == "True Positive graphs":
                        if tp_graphs:
                            graph_viewer("TP", tp_graphs, tp_sentences, nodes)

                    elif tp_fp_fn == "False Negative graphs":
                        if fn_graphs:
                            graph_viewer("FN", fn_graphs, fn_sentences, nodes)


def get_args():
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("-t", "--train-data", type=str, required=True)
    parser.add_argument("-v", "--val-data", type=str)
    parser.add_argument(
        "-sr",
        "--suggested-rules",
        default=None,
        type=str,
        help="Rules extracted automatically from python. If not present, the UI will automatically train it.",
    )
    parser.add_argument(
        "-hr",
        "--hand-rules",
        default=None,
        type=str,
        help="Rules extracted with the UI. If provided, the UI will load them.",
    )
    parser.add_argument("-m", "--mode", default="supervised", type=str)
    parser.add_argument("-g", "--graph-format", default="fourlang", type=str)
    return parser.parse_args()


def main(args):
    st.set_page_config(layout="wide")
    st.markdown(
        "<h1 style='text-align: center; color: black;'>Rule extraction framework</h1>",
        unsafe_allow_html=True,
    )

    init_session_states()
    evaluator = init_evaluator()
    data = read_train(args.train_data)
    if args.val_data:
        val_data = read_val(args.val_data)
    graph_format = args.graph_format
    feature_path = args.suggested_rules
    hand_made_rules = args.hand_rules
    mode = args.mode
    if mode == "supervised":
        assert args.val_data
        supervised_mode(
            evaluator, data, val_data, graph_format, feature_path, hand_made_rules
        )
    elif mode == "unsupervised":
        unsupervised_mode(
            evaluator, args.train_data, graph_format, feature_path, hand_made_rules
        )


if __name__ == "__main__":
    args = get_args()
    main(args)



import copy
import json
import re
import sys

import pandas as pd
import streamlit as st
from streamlit.report_thread import REPORT_CONTEXT_ATTR_NAME

from xpotato.dataset.utils import default_pn_to_graph
from xpotato.graph_extractor.extract import FeatureEvaluator
from xpotato.models.trainer import GraphTrainer

from contextlib import contextmanager
from io import StringIO
from threading import current_thread


def init_session_states():
    if "false_graph_number" not in st.session_state:
        st.session_state.false_graph_number = 0
    if "true_graph_number" not in st.session_state:
        st.session_state.true_graph_number = 0
    if "false_neg_number" not in st.session_state:
        st.session_state.false_neg_number = 0
    if "predicted_num" not in st.session_state:
        st.session_state.predicted_num = 0
    if "whole_accuracy" not in st.session_state:
        st.session_state.whole_accuracy = []
    if "df_statistics" not in st.session_state:
        st.session_state.df_statistics = pd.DataFrame
    if "val_dataframe" not in st.session_state:
        st.session_state.val_dataframe = pd.DataFrame
    if "whole_accuracy_val" not in st.session_state:
        st.session_state.whole_accuracy_val = []
    if "feature_df" not in st.session_state:
        st.session_state.feature_df = pd.DataFrame
    if "clustered_words_path" not in st.session_state:
        st.session_state.clustered_words_path = None
    if "features" not in st.session_state:
        st.session_state.features = {}
    if "suggested_features" not in st.session_state:
        st.session_state.suggested_features = {}
    if "trained" not in st.session_state:
        st.session_state.trained = False
    if "ml_feature" not in st.session_state:
        st.session_state.ml_feature = None
    if "sens" not in st.session_state:
        st.session_state.sens = []
    if "min_edge" not in st.session_state:
        st.session_state.min_edge = 0

    if "rows_to_delete" not in st.session_state:
        st.session_state.rows_to_delete = []
    if "rls_after_delete" not in st.session_state:
        st.session_state.rls_after_delete = []

    if "df_to_train" not in st.session_state:
        st.session_state.df_to_train = pd.DataFrame

    if "applied_rules" not in st.session_state:
        st.session_state.applied_rules = []

    if "rank" not in st.session_state:
        st.session_state.rank = False


def rerun():
    raise st.experimental_rerun()


@contextmanager
def st_redirect(src, dst):
    placeholder = st.empty()
    output_func = getattr(placeholder, dst)

    with StringIO() as buffer:
        old_write = src.write

        def new_write(b):
            if getattr(current_thread(), REPORT_CONTEXT_ATTR_NAME, None):
                buffer.write(b)
                output_func(buffer.getvalue())
            else:
                old_write(b)

        try:
            src.write = new_write
            yield
        finally:
            src.write = old_write


@contextmanager
def st_stdout(dst):
    with st_redirect(sys.stdout, dst):
        yield


def to_dot(graph, marked_nodes=set(), integ=False):
    lines = ["digraph finite_state_machine {", "\tdpi=70;"]
    # lines.append('\tordering=out;')
    # sorting everything to make the process deterministic
    node_lines = []
    node_to_name = {}
    for node, n_data in graph.nodes(data=True):
        if integ:
            d_node = d_clean(str(node))
        else:
            d_node = d_clean(n_data["name"]) if n_data["name"] else "None"
        printname = d_node
        node_to_name[node] = printname
        if "expanded" in n_data and n_data["expanded"] and printname in marked_nodes:
            node_line = '\t{0} [shape = circle, label = "{1}", \
                    style=filled, fillcolor=purple];'.format(
                d_node, printname
            ).replace(
                "-", "_"
            )
        elif "expanded" in n_data and n_data["expanded"]:
            node_line = '\t{0} [shape = circle, label = "{1}", \
                    style="filled"];'.format(
                d_node, printname
            ).replace(
                "-", "_"
            )
        elif "fourlang" in n_data and n_data["fourlang"]:
            node_line = '\t{0} [shape = circle, label = "{1}", \
                    style="filled", fillcolor=red];'.format(
                d_node, printname
            ).replace(
                "-", "_"
            )
        elif "substituted" in n_data and n_data["substituted"]:
            node_line = '\t{0} [shape = circle, label = "{1}", \
                    style="filled"];'.format(
                d_node, printname
            ).replace(
                "-", "_"
            )
        elif printname in marked_nodes:
            node_line = '\t{0} [shape = circle, label = "{1}", style=filled, fillcolor=lightblue];'.format(
                d_node, printname
            ).replace(
                "-", "_"
            )
        else:
            node_line = '\t{0} [shape = circle, label = "{1}"];'.format(
                d_node, printname
            ).replace("-", "_")
        node_lines.append(node_line)
    lines += sorted(node_lines)

    edge_lines = []
    for u, v, edata in graph.edges(data=True):
        if "color" in edata:
            d_node1 = node_to_name[u].replace("-", "_")
            d_node2 = node_to_name[v].replace("-", "_")
            edge_lines.append(
                '\t{0} -> {1} [ label = "{2}" ];'.format(
                    d_node1, d_node2, edata["color"]
                )
            )

    lines += sorted(edge_lines)
    lines.append("}")
    return "\n".join(lines)


def save_ruleset(path, features):
    with open(path, "w+") as f:
        json.dump(features, f)


def d_clean(string):
    s = string
    for c in "\\=@-,'\".!:;<>/{}[]()#^?":
        s = s.replace(c, "_")
    s = s.replace("$", "_dollars")
    s = s.replace("%", "_percent")
    s = s.replace("|", " ")
    s = s.replace("*", " ")
    if s == "#":
        s = "_number"
    keywords = ("graph", "node", "strict", "edge")
    if re.match("^[0-9]", s) or s in keywords:
        s = "X" + s
    return s


def get_df_from_rules(rules, negated_rules):
    data = {"rules": rules, "negated_rules": negated_rules}
    # Create DataFrame.
    df = pd.DataFrame(data)

    return df


def save_after_modify(hand_made_rules, classes):
    st.session_state.features[classes] = copy.deepcopy(
        st.session_state.rls_after_delete
    )
    st.session_state.feature_df = get_df_from_rules(
        [";".join(feat[0]) for feat in st.session_state.features[classes]],
        [";".join(feat[1]) for feat in st.session_state.features[classes]],
    )
    save_rules = hand_made_rules or "saved_features.json"
    save_ruleset(save_rules, st.session_state.features)
    st.session_state.rows_to_delete = []
    rerun()


@st.cache(allow_output_mutation=True)
def load_text_to_4lang():
    tfl = TextTo4lang("en", "en_nlp_cache")
    return tfl


@st.cache()
def init_evaluator():
    return FeatureEvaluator()


@st.cache(allow_output_mutation=True)
def read_train(path):
    return pd.read_pickle(path)


def save_dataframe(data, path):
    data.to_pickle(path)


@st.cache(allow_output_mutation=True)
def read_val(path):
    return pd.read_pickle(path)


def train_df(df, min_edge=0, rank=False):
    with st_stdout("code"):
        trainer = GraphTrainer(df)
        features = trainer.prepare_and_train(min_edge=min_edge, rank=rank)

        return features


def rule_chooser():
    option = st.selectbox("Choose from the rules", st.session_state.sens)
    G, _ = default_pn_to_graph(option.split(";")[0])
    text_G, _ = default_pn_to_graph(option.split(";")[0])
    st.graphviz_chart(to_dot(text_G), use_container_width=True)
    nodes = [d_clean(n[1]["name"].split("_")[0]) for n in text_G.nodes(data=True)]
    return nodes, option


def annotate_df(predicted):
    for i, pred in enumerate(predicted):
        if pred:
            st.session_state.df.at[i, "label"] = st.session_state.inverse_labels[1]
            st.session_state.df.at[i, "applied_rules"] = pred
        else:
            st.session_state.df.at[i, "applied_rules"] = []
            if st.session_state.df.loc[i, "annotated"] == False:
                st.session_state.df.at[i, "label"] = ""


def show_ml_feature(classes, hand_made_rules):
    st.markdown(
        f"<span>Feature: {st.session_state.ml_feature[0]}, Retrieved True Positive samples: <b>{st.session_state.ml_feature[5]}</b>, \
                        Retrieved False Positive samples: <b>{st.session_state.ml_feature[6]}</b></span>",
        unsafe_allow_html=True,
    )
    accept_rule = st.button("Accept")
    decline_rule = st.button("Decline")
    if accept_rule:
        st.session_state.features[classes].append(st.session_state.ml_feature[0])
        st.session_state.ml_feature = None
        if st.session_state.features[classes]:
            st.session_state.feature_df = get_df_from_rules(
                [";".join(feat[0]) for feat in st.session_state.features[classes]],
                [";".join(feat[1]) for feat in st.session_state.features[classes]],
            )
            save_rules = hand_made_rules or "saved_features.json"
            save_ruleset(save_rules, st.session_state.features)
            rerun()
    elif decline_rule:
        st.session_state.ml_feature = None
        rerun()


def extract_data_from_dataframe(option):
    fp_graphs = st.session_state.df_statistics.iloc[
        st.session_state.sens.index(option)
    ].False_positive_graphs
    fp_sentences = st.session_state.df_statistics.iloc[
        st.session_state.sens.index(option)
    ].False_positive_sens
    tp_graphs = st.session_state.df_statistics.iloc[
        st.session_state.sens.index(option)
    ].True_positive_graphs
    tp_sentences = st.session_state.df_statistics.iloc[
        st.session_state.sens.index(option)
    ].True_positive_sens
    fn_graphs = st.session_state.df_statistics.iloc[
        st.session_state.sens.index(option)
    ].False_negative_graphs
    fn_sentences = st.session_state.df_statistics.iloc[
        st.session_state.sens.index(option)
    ].False_negative_sens
    prec = st.session_state.df_statistics.iloc[
        st.session_state.sens.index(option)
    ].Precision
    recall = st.session_state.df_statistics.iloc[
        st.session_state.sens.index(option)
    ].Recall
    fscore = st.session_state.df_statistics.iloc[
        st.session_state.sens.index(option)
    ].Fscore
    support = st.session_state.df_statistics.iloc[
        st.session_state.sens.index(option)
    ].Support
    predicted = st.session_state.df_statistics.iloc[
        st.session_state.sens.index(option)
    ].Predicted
    return (
        fn_graphs,
        fn_sentences,
        fp_graphs,
        fp_sentences,
        fscore,
        prec,
        predicted,
        recall,
        support,
        tp_graphs,
        tp_sentences,
    )


def graph_viewer(type, graphs, sentences, nodes):

    graph_type = {
        "FP": st.session_state.false_graph_number,
        "TP": st.session_state.true_graph_number,
        "FN": st.session_state.false_neg_number,
    }
    if st.button(f"Previous {type}"):
        graph_type[type] = max(0, graph_type[type] - 1)
    if st.button(f"Next {type}"):
        graph_type[type] = min(
            graph_type[type] + 1,
            len(graphs) - 1,
        )
    if graph_type[type] > len(graphs) - 1:
        graph_type[type] = 0
    st.markdown(
        f"<span><b>Sentence:</b> {sentences[graph_type[type]][0]}</span>",
        unsafe_allow_html=True,
    )
    st.markdown(
        f"<span><b>Gold label:</b> {sentences[graph_type[type]][1]}</span>",
        unsafe_allow_html=True,
    )
    st.text(f"{type}: {len(graphs)}")
    current_graph = graphs[graph_type[type]]
    st.graphviz_chart(
        to_dot(
            current_graph,
            marked_nodes=set(nodes),
        ),
        use_container_width=True,
    )

    if type == "FP":
        st.session_state.false_graph_number = graph_type[type]
    elif type == "TP":
        st.session_state.true_graph_number = graph_type[type]
    elif type == "FN":
        st.session_state.false_neg_number = graph_type[type]


def add_rule_manually(classes, hand_made_rules):
    text = st.text_area("You can add a new rule here manually")
    negated_text = st.text_area("You can modify the negated features here")
    agree = st.button("Add rule to the ruleset")
    if agree:
        if not negated_text.strip():
            negated_features = []
        else:
            negated_features = negated_text.split(";")
        st.session_state.features[classes].append([[text], negated_features, classes])
        if st.session_state.features[classes]:
            st.session_state.feature_df = get_df_from_rules(
                [";".join(feat[0]) for feat in st.session_state.features[classes]],
                [";".join(feat[1]) for feat in st.session_state.features[classes]],
            )
            save_rules = hand_made_rules or "saved_features.json"
            save_ruleset(save_rules, st.session_state.features)
            rerun()
    st.markdown(
        f"<span><b>Or get suggestions by our ML!</b></span>",
        unsafe_allow_html=True,
    )


def rank_and_suggest(classes, data, evaluator, rank_false_negatives=True):
    suggest_new_rule = st.button("suggest new rules")
    if suggest_new_rule:
        if (
            not st.session_state.df_statistics.empty
            and st.session_state.sens
            and st.session_state.suggested_features[classes]
        ):
            features_to_rank = st.session_state.suggested_features[classes][:5]
            with st.spinner("Ranking rules..."):
                features_ranked = evaluator.rank_features(
                    classes,
                    features_to_rank,
                    data,
                    st.session_state.df_statistics.iloc[0].False_negative_indices,
                )
            suggested_feature = features_ranked[0]

            st.session_state.suggested_features[classes].remove(suggested_feature[0])

            st.session_state.ml_feature = suggested_feature
        else:
            st.warning("Dataset is not evaluated!")
